/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */


#include <windows.h>
#include <malloc.h>

#include "log.h"

#include "constants.h"
#include "FileCache2.h"

CFileCache2::CFileCache2(void)
{
	m_hFile = INVALID_HANDLE_VALUE;
	m_qwTotalFileSize = (QWORD)0;
	m_qwFilePos = (QWORD)0;

	memset(&m_bfr, 0, sizeof(m_bfr));
}

BOOL CFileCache2::Init(HANDLE hFile, DWORD dwLog2Size)
{
	BOOL bRtnVal = FALSE;
	DWORD dwFileSizeHigh = 0, dwFileSizeLow = 0;

	m_hFile = hFile;
	m_bfr.dwLog2Size = dwLog2Size;
	m_bfr.dwBfrSize = 0x1 << m_bfr.dwLog2Size;
	m_bfr.dwBfrMask = m_bfr.dwBfrSize -1;
	m_bfr.dwBfrReloadGap = m_bfr.dwBfrSize/8;
	m_bfr.dwBfrUsableSize = m_bfr.dwBfrSize - m_bfr.dwBfrReloadGap;

	if (IS_VALID_HANDLE(hFile))
	{
		// Don't use GetFileSizeEx, it's incompatible with older OS's. This is just as good.
		dwFileSizeLow = ::GetFileSize(m_hFile, &dwFileSizeHigh);

		if (dwFileSizeLow != INVALID_FILE_SIZE)
		{
			m_qwTotalFileSize = ((QWORD)dwFileSizeHigh << 32);
			m_qwTotalFileSize = m_qwTotalFileSize + ((QWORD)(dwFileSizeLow));

			// the user is allow to init this object even if he's already read from this file
			// "manually", so we init our local file pos to the real one.
			m_qwFilePos = GetDiskFilePos();

			bRtnVal = TRUE;
		}
	}
	// OK, now malloc the buffer if we made it this far
	if (bRtnVal)
		m_bfr.pRingBuf = (BYTE*) malloc(m_bfr.dwBfrSize);

	return (bRtnVal && m_bfr.pRingBuf != NULL);
}

// This is called constantly, and simply returns as long as the number of bytes in the
// buffer is equal to or greater than count. When the buffer finally runs low enough
// said condition fails, the buffer is reloaded to 7/8 full (unrelated to the value of
// count - count is basically just used to determine "reload" or "no reload necessary").
//
BOOL CFileCache2::BytesNeeded(DWORD dwCount)
{
	BOOL		bRtnVal = TRUE;
	DWORD		dwBytesToReload, dwBytesLeft, dwTempByteCount, dwBytesRead1, dwBytesRead2 = 0;

	if (BytesInCache() < dwCount)	// if this isn't true, nothing need be done
	{
		// reload required: refill the buffer so it's 7/8 full again. First determine
		// how many bytes that is. Then use standard ring buffer reload logic
		dwBytesToReload  = m_bfr.dwBfrUsableSize - BytesInCache();	// m_bfr.dwBfrUsableSize is around 7/8 of the max
		dwBytesLeft = dwBytesToReload;

		if (m_bfr.dwHeadIdx < m_bfr.dwTailIdx)
			ReadFile(m_hFile, &m_bfr.pRingBuf[m_bfr.dwHeadIdx], dwBytesLeft, &dwBytesRead1, NULL);

		else
		{
			// Else one or possibly two copies is necessary
			dwTempByteCount = min(dwBytesLeft, m_bfr.dwBfrSize - m_bfr.dwHeadIdx);
			ReadFile(m_hFile, &m_bfr.pRingBuf[m_bfr.dwHeadIdx], dwTempByteCount, &dwBytesRead1, NULL);
			dwBytesLeft -= dwTempByteCount;

			// if there are still more bytes, they go at the beginning of the buffer
			if (dwBytesLeft > 0)
				ReadFile(m_hFile, m_bfr.pRingBuf, dwBytesLeft, &dwBytesRead2, NULL);
		}
		// either way, advance the head by the total number of bytes requested
		m_bfr.dwHeadIdx = (m_bfr.dwHeadIdx + dwBytesToReload) & m_bfr.dwBfrMask;

		// if #of bytes requested differs from number actually read, it was an EOF condition & we return FAIL
		bRtnVal = (dwBytesRead1 + dwBytesRead2 <= dwBytesToReload);
	}
	_ASSERT_(bRtnVal);
	return(bRtnVal);
}

// User call to get bytes from the buffer. Always calls BytesNeeded() first, to make sure there
// are at least as many bytes in the buffer as requested. Then performs the transfer with either
// one or two copies, depending on the head and tail pointers.
DWORD CFileCache2::ReadBytes(void* pBuf, DWORD dwCount)
{
	DWORD	dwRtnVal = 0;
	DWORD	dwTempByteCount;
	DWORD	dwBytesLeft = dwCount;

	if (dwBytesLeft < m_bfr.dwBfrUsableSize)			// request must be legal
	{
		BytesNeeded(dwBytesLeft);						// refill buffer if needed

		// perform the copy if a non-null target buffer was specified
		if (pBuf != NULL)
		{
			if (m_bfr.dwTailIdx < m_bfr.dwHeadIdx)	// only one copy required
				memcpy(pBuf, &m_bfr.pRingBuf[m_bfr.dwTailIdx], dwBytesLeft);

			else
			{
				// may require one or two copies
				dwTempByteCount = min(dwBytesLeft, m_bfr.dwBfrSize - m_bfr.dwTailIdx);
				memcpy(pBuf, &m_bfr.pRingBuf[m_bfr.dwTailIdx], dwTempByteCount);
				dwBytesLeft -= dwTempByteCount;

				// if there are still more bytes, they come from the beginning of the buffer
				if (dwBytesLeft > 0)
					memcpy(&((char*)pBuf)[dwTempByteCount], m_bfr.pRingBuf, dwBytesLeft);
			}
		}
		// either way, advance the tail by the total number of bytes requested
		m_bfr.dwTailIdx = (m_bfr.dwTailIdx + dwCount) & m_bfr.dwBfrMask;
		m_qwFilePos += dwCount;		// continuous running file pointer points to the absolute next file offset
		//m_dwFilePos += dwCount;
		dwRtnVal = dwCount;		// return value is always what was requested, unless init condition failed
	}
	return(dwCount);
}

// same as ReadBytes with a null target buffer specified, but function can also be used for clarity
DWORD CFileCache2::SkipBytes(DWORD dwCount)
{
	return(ReadBytes(NULL, dwCount));
}

CFileCache2::~CFileCache2(void)
{
	if (m_bfr.pRingBuf != NULL)
		free(m_bfr.pRingBuf);
}

QWORD CFileCache2::GetFileSize()
{
	return(m_qwTotalFileSize);
}

// These are LE functions coded to operate on an LE O/S. Adding any/all other
// other variations is a rather trivial exercise, but this particular version
// of the source includes only the functions required for running on Windows (LE)
// and reading ASF format files (all data is LE).
//
// Note to the curious: The ASF file format is actually more the exception than
// the rule in this regard; many or most other formats store their multi-byte
// values in Big Endian (Network) byte order. But this release, for the most part,
// contains only the functions needed for the job at hand.

QWORD CFileCache2::NextVarlenLe(DWORD dwBytes)
{
	BYTE	uchBuf[8];
	QWORD	qwResult = 0;

	_ASSERT_ (dwBytes <= 8);	// 64-bit max supported

	ReadBytes(uchBuf, dwBytes);

	for (int nIdx = dwBytes-1; nIdx >= 0; nIdx--)
	{
		qwResult *= 256;
		qwResult += uchBuf[nIdx];
	}
	return(qwResult);
}

// Variable length code specific to ASF files
DWORD CFileCache2::NextAsfVarlen(DWORD dwLenCode)
{
	DWORD dwResult = 0;

	switch(dwLenCode)
	{
	  case 3:
	    dwResult = NextDWordLe();
		break;
	  case 2:
		dwResult = NextWordLe();
		break;
	  case 1:
		dwResult = NextByte();
		break;
	  default:
		break;
	}
	return(dwResult);
}

QWORD CFileCache2::NextQWordLe(void)
{
	BYTE	uchBuf[sizeof(QWORD)];

	ReadBytes(uchBuf, sizeof(QWORD));
	return(*((QWORD*)uchBuf));
}

DWORD CFileCache2::NextDWordLe(void)
{
	BYTE	uchBuf[sizeof(DWORD)];

	ReadBytes(uchBuf, sizeof(DWORD));
	return(*((DWORD*)uchBuf));
}

WORD CFileCache2::NextWordLe(void)
{
	BYTE	uchBuf[sizeof(WORD)];

	ReadBytes(uchBuf, sizeof(WORD));
	return(*((WORD*)uchBuf));
}

BYTE CFileCache2::NextByte(void)
{
	BYTE	uchByte;

	ReadBytes(&uchByte, sizeof(BYTE));
	return(uchByte);
}

void CFileCache2::GetNextGuid(GUID* pGuid)
{
	ReadBytes(pGuid, sizeof(GUID));
}

QWORD CFileCache2::GetFilePos(void)
{
	return(m_qwFilePos);
}

QWORD CFileCache2::GetDiskFilePos(void)
{
	DWORD	dwPosLow;
	long	nPosHi = 0;
	QWORD	qwRtnVal;

	// presumably we don't have to worry about the file being so large that the sign bit
	// of nPosHi becomes an issue. The file size would have to exceed 2^63!
	dwPosLow = SetFilePointer(m_hFile, 0, &nPosHi, FILE_CURRENT);

	qwRtnVal = ((QWORD)nPosHi << 32) + (QWORD)dwPosLow;

	return(qwRtnVal);
}