/* *******************************************************************************************
The MIT License

Copyright (c) 2007 Steven G. Greenberg

Permission is hereby granted, free of charge, to any person obtaining a copy of this
software and associated documentation files (the "Software"), to deal in the Software
without restriction, including without limitation the rights to use, copy, modify, merge,
publish, distribute, sublicense, and/or sell copies of the Software, and to permit persons
to whom the Software is furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all copies or
substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE
FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR
OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
DEALINGS IN THE SOFTWARE.
******************************************************************************************* */

#pragma pack(1)		// important!

struct GUID_TO_NAME_STRING	// kinda basic way to build a naming table for guids
{
	const GUID*	pGuid;
	const char*	pszName;
};

// ======================= ASF file format related structures ==========================
struct ASF_ID_AND_SIZE
{
	GUID				guidId;
	QWORD	qwSize;
};

struct ASF_MAIN_HDR_OBJ
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	DWORD					dwNumAddlHdrObjs;
	BYTE					uchReserved1;
	BYTE					uchReserved2;
};

struct ASF_VIDEO_MEDIA_OBJECT
{
	DWORD					dwEncoded_Image_Width;
	DWORD					dwEncoded_Image_Height;
	BYTE					uchReserved;
	WORD					wFormat_Data_Size;
	BITMAPINFOHEADER		biInfo;
	BYTE					uchUnknown;		// this mystery byte makes it all work out
	// variable length specialized data follows - what we want!
};

struct STREAM_PROPERTIES_OBJECT
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	GUID					guidStream_Type;
	GUID					guidError_Correction_Type;
	QWORD					qwTime_Offset;
	DWORD					dwType_Specific_Data_Length;
	DWORD					dwError_Correction_Data_Length;
	WORD					wFlags;					// 7 ls bits are stream number. MSB means encrypted
	DWORD					dwReserved;
// variable length Type_Specific_Data follows, depending on stream type
// another variable length STREAM_PROPERTIES_OBJECT member follows
// concerning error correction, which we have no interest in
};

struct ASF_HEADER_EXTENSION_OBJECT
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	GUID					guidReserved1;
	WORD					wReserved2;
	DWORD					dwHeaderExtensionSize;
	// variable length data starts here
};

struct FILE_PROPERTIES_OBJECT
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	GUID				guidFile_ID;
	QWORD				qwFile_Size;
	QWORD				qwCreation_Date;
	QWORD				qwData_Packets_Count;
	QWORD				qwPlay_Duration;
	QWORD				qwSend_Duration;
	QWORD				qwPreroll;
	DWORD				dwFlags;
	DWORD				Minimum_Data_Packet_Size;
	DWORD				Maximum_Data_Packet_Size;
	DWORD				Maximum_Bitrate;
};

struct EXTENDED_STREAM_PROPERTIES_OBJECT
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	QWORD					qwStartTime;
	QWORD					qwEndTime;
	DWORD					dwDataBitrate;
	DWORD					dwBufferSize;
	DWORD					dwInitialBufferFullness;
	DWORD					dwAlternateDataBitrate;
	DWORD					dwAlternateBufferSize;
	DWORD					dwAlternateInitialBufferFullness;
	DWORD					dwMaximumObjectSize;
	DWORD					dwFlags;
	WORD					dwStreamNumber;
	WORD					dwStreamLanguageIdIndex;
	QWORD					qwAverageTimePerFrame;
	WORD					wStreamNameCount;
	WORD					wPayloadExtensionSystemCount;
// variable length Stream Names	follows
// variable length Payload Extension Systems follows
};

struct METADATA_OBJECT
{
	struct ASF_ID_AND_SIZE	asfIdAndSize;
	WORD	wRecordCount;
	// variable data follows
};

struct ASF_METADATA_DESCRIPTION_RECORD
{
	WORD	wReserved;		// must be zero
	WORD	wStreamNumber;
	WORD	wNameLength;
	WORD	wDataType;
	DWORD	dwDataLength;
	// WCHAR	Name[1];	// variable length name
	// variable length data follows variable length word
};

// ======================= VC-1 related structures  ==========================

struct VC1_HEADER_SEQUENCE
{
	BYTE		uchByteSeq[MAX_SEQ_PLUS_ENTRY_LEN];
	DWORD		dwLength;
};

// our structure used for a constant framerate lookup table (numerators & denominators vs rates)
struct PERIOD_TO_NR_DR
{
	DWORD	dwPeriodMicroSecs;
	DWORD	dwNumIdx;
	DWORD	dwDenIdx;
};

// a structure used to define a global table that holds up to 32 pairs of these
struct LEAKY_BUCKET
{
	WORD	wHardRate;
	WORD	wHardBuffer;
};

// =========== "Variable Bit Length Conditional Parsing Automaton" structure ===========
//
// The structure defines a parsing mechanism, complete with conditionals and "jumps".
// Indexing thru a table of these entries comprises the bulk of the VC-1 parsing logic.
// The original idea of the table was for it to contain only pre-initialized constant
// values representing the syntax to be parsed. The entries get "executed" almost as
// an interpreted language would execute its code. But for simplicity, the table is
// currently a mix of "instructions and data" - as can be seen below, each entry also
// contains the actual value of the data item it's associated with (and a flag indicating
// if that value is valid). Those get filled in at run time; if the table is running in
// "parse" mode, the value gets filled in from the value derived by parsing the bits. In
// the reverse mode, the value is "manually" set at run time before the interpreter is
// started, and the value therein controls the contents of the bitstream that comes out.
//

struct BIT_PARSE_TABLE
{
	enum VC1_HDR_PARAMS		nParam;				// pre-initialized const val
	const char*				pszParamName;		// pre-initialized const val
	DWORD					dwNumBits;			// pre-initialized const val
	BOOL					bValueIsValid;		// variable - TRUE if value below is valid
	DWORD					dwValue;			// variable - a value associated with this entry
	enum VC1_HDR_PARAMS		nParamToCompare;	// pre-initialized const val
	enum COMPARE_TYPE		nCompareType;		// pre-initialized const val
	DWORD					dwCompareTargVal;	// pre-initialized const val
	enum VC1_HDR_PARAMS		nParamToSkipTo;		// pre-initialized const val
};

struct OPTION
{
	enum CMND_LINE_FLAG	nFlag;
	const wchar_t*		wpszShortName;
	const wchar_t*		wpszLongName;
	BOOL				bFlagVal;
};
